from math import ceil
from typing import Optional

from riot_na.airr.airr_validator import validate_airr_entry
from riot_na.alignment.alignment_utils import fold_cigar, has_frameshift
from riot_na.data.model import (
    AirrRearrangementEntryNT,
    AlignmentEntryNT,
    AlignmentString,
    RegionOffsetsAA,
    RegionOffsetsNT,
    Scheme,
    SchemeAlignment,
)


class AirrBuilder:  # pylint: disable=too-many-instance-attributes
    def __init__(self, sequence_header: str, sequence: str, scheme: Scheme):
        self.sequence_header = sequence_header
        self.sequence = sequence
        self.scheme = scheme

        self.rearrangement = AirrRearrangementEntryNT(
            sequence_header=sequence_header, sequence=sequence, numbering_scheme=scheme.value
        )

        self.v_gene_sequence: Optional[str] = None
        self.v_gene_alignment: Optional[AlignmentEntryNT] = None

        self.j_gene_alignment: Optional[AlignmentEntryNT] = None
        self.j_gene_sequence: Optional[str] = None

        self.c_gene_alignment: Optional[AlignmentEntryNT] = None
        self.c_gene_sequence: Optional[str] = None
        self.c_gene_name: Optional[str] = None

        self.d_gene_alignment: Optional[AlignmentEntryNT] = None
        self.d_gene_sequence: Optional[str] = None
        self.d_gene_name: Optional[str] = None

        self.scheme_alingment: Optional[SchemeAlignment] = None
        self.nt_offsets: Optional[RegionOffsetsNT] = None
        self.aa_offsets: Optional[RegionOffsetsAA] = None

    def with_v_gene_alignment(
        self,
        v_aln: AlignmentEntryNT,
    ):
        self.v_gene_sequence = v_aln.t_seq
        self.rearrangement.rev_comp = v_aln.rev_comp
        self.v_gene_alignment = v_aln

        if v_aln is None:
            return self

        self.rearrangement.v_call = v_aln.target_id
        self.rearrangement.locus = v_aln.locus.value

        # alignment positions are 0-indexed, add 1 to start index to convert to 1-based
        self.rearrangement.v_sequence_start = v_aln.q_start + 1
        self.rearrangement.v_sequence_end = v_aln.q_end

        self.rearrangement.v_germline_start = v_aln.t_start + 1
        self.rearrangement.v_germline_end = v_aln.t_end

        self.rearrangement.v_sequence_alignment = self.sequence[v_aln.q_start : v_aln.q_end]

        # 1 based, hence need to add 1
        self.rearrangement.v_alignment_start = 1
        self.rearrangement.v_alignment_end = v_aln.q_end - v_aln.q_start

        self.rearrangement.v_germline_alignment = v_aln.t_seq[v_aln.t_start : v_aln.t_end]

        self.rearrangement.v_score = v_aln.alignment_score
        self.rearrangement.v_cigar = v_aln.cigar
        self.rearrangement.v_support = v_aln.e_value
        self.rearrangement.v_identity = v_aln.seq_identity

        return self

    def with_j_gene_alignment(
        self,
        j_aln: AlignmentEntryNT,
    ):
        self.j_gene_sequence = j_aln.t_seq

        if j_aln is not None:
            assert self.v_gene_alignment is not None

            self.j_gene_alignment = j_aln

            self.rearrangement.j_call = j_aln.target_id
            # alignment positions are 0-indexed, add 1 to start index to convert to 1-based
            self.rearrangement.j_sequence_start = j_aln.q_start + 1
            self.rearrangement.j_sequence_end = j_aln.q_end

            self.rearrangement.j_germline_start = j_aln.t_start + 1
            self.rearrangement.j_germline_end = j_aln.t_end

            self.rearrangement.j_sequence_alignment = self.sequence[j_aln.q_start : j_aln.q_end]

            # 1 based, hence need to add 1
            assert self.rearrangement.v_sequence_start is not None
            self.rearrangement.j_alignment_start = j_aln.q_start + 1 - self.v_gene_alignment.q_start
            self.rearrangement.j_alignment_end = j_aln.q_end - self.v_gene_alignment.q_start

            self.rearrangement.j_germline_alignment = j_aln.t_seq[j_aln.t_start : j_aln.t_end]

            self.rearrangement.j_score = j_aln.alignment_score
            self.rearrangement.j_cigar = j_aln.cigar
            self.rearrangement.j_support = j_aln.e_value
            self.rearrangement.j_identity = j_aln.seq_identity

            assert j_aln.reading_frame is not None
            self.rearrangement.j_frame = (3 - ((j_aln.t_start - j_aln.reading_frame) % 3)) % 3

        return self

    def with_c_gene_alignment(
        self,
        c_aln: AlignmentEntryNT,
    ):
        self.c_gene_sequence = c_aln.t_seq
        self.c_gene_name = c_aln.target_id

        if self.j_gene_alignment is None:
            return self

        self.c_gene_alignment = c_aln

        self.rearrangement.c_call = c_aln.target_id
        # alignment positions are 0-indexed
        self.rearrangement.c_sequence_start = c_aln.q_start + 1
        self.rearrangement.c_sequence_end = c_aln.q_end

        self.rearrangement.c_germline_start = c_aln.t_start + 1
        self.rearrangement.c_germline_end = c_aln.t_end

        self.rearrangement.c_sequence_alignment = self.sequence[c_aln.q_start : c_aln.q_end]

        # 1 based, hence need to add 1
        assert self.v_gene_alignment is not None
        self.rearrangement.c_alignment_start = c_aln.q_start + 1 - self.v_gene_alignment.q_start
        self.rearrangement.c_alignment_end = c_aln.q_end - self.v_gene_alignment.q_start

        self.rearrangement.c_germline_alignment = c_aln.t_seq[c_aln.t_start : c_aln.t_end]

        self.rearrangement.c_score = c_aln.alignment_score
        self.rearrangement.c_cigar = c_aln.cigar
        self.rearrangement.c_support = c_aln.e_value
        self.rearrangement.c_identity = c_aln.seq_identity

        return self

    def with_d_gene_alignment(
        self,
        d_aln: AlignmentEntryNT,
    ):
        self.d_gene_sequence = d_aln.t_seq
        self.d_gene_name = d_aln.target_id

        if self.j_gene_alignment is None or self.v_gene_alignment is None:
            return self

        self.d_gene_alignment = d_aln

        self.rearrangement.d_call = d_aln.target_id
        # alignment positions are 0-indexed
        self.rearrangement.d_sequence_start = d_aln.q_start + 1
        self.rearrangement.d_sequence_end = d_aln.q_end

        self.rearrangement.d_germline_start = d_aln.t_start + 1
        self.rearrangement.d_germline_end = d_aln.t_end

        self.rearrangement.d_sequence_alignment = self.sequence[d_aln.q_start : d_aln.q_end]

        assert self.v_gene_alignment is not None
        self.rearrangement.d_alignment_start = d_aln.q_start + 1 - self.v_gene_alignment.q_start
        self.rearrangement.d_alignment_end = d_aln.q_end - self.v_gene_alignment.q_start

        self.rearrangement.d_germline_alignment = d_aln.t_seq[d_aln.t_start : d_aln.t_end]

        self.rearrangement.d_score = d_aln.alignment_score
        self.rearrangement.d_cigar = d_aln.cigar
        self.rearrangement.d_support = d_aln.e_value
        self.rearrangement.d_identity = d_aln.seq_identity

        return self

    def with_sequence_alignment_aa(self, aligned_sequence_aa: Optional[str]):
        self.rearrangement.sequence_alignment_aa = aligned_sequence_aa
        return self

    def with_sequence_alignment(self, aligned_sequence: str):
        self.rearrangement.sequence_alignment = aligned_sequence
        return self

    def with_reading_frame(self, reading_frame: int):
        self.rearrangement.v_frame = reading_frame
        return self

    def with_nt_region_offsets(self, offsets: RegionOffsetsNT):
        self.nt_offsets = offsets

        self.rearrangement.fwr1_start = offsets.fwr1_start
        self.rearrangement.fwr1_end = offsets.fwr1_end
        self.rearrangement.cdr1_start = offsets.cdr1_start
        self.rearrangement.cdr1_end = offsets.cdr1_end
        self.rearrangement.fwr2_start = offsets.fwr2_start
        self.rearrangement.fwr2_end = offsets.fwr2_end
        self.rearrangement.cdr2_start = offsets.cdr2_start
        self.rearrangement.cdr2_end = offsets.cdr2_end
        self.rearrangement.fwr3_start = offsets.fwr3_start
        self.rearrangement.fwr3_end = offsets.fwr3_end
        self.rearrangement.cdr3_start = offsets.cdr3_start
        self.rearrangement.cdr3_end = offsets.cdr3_end
        self.rearrangement.fwr4_start = offsets.fwr4_start
        self.rearrangement.fwr4_end = offsets.fwr4_end

        self.rearrangement.fwr1 = self._extract_region(offsets.fwr1_start, offsets.fwr1_end)
        self.rearrangement.cdr1 = self._extract_region(offsets.cdr1_start, offsets.cdr1_end)
        self.rearrangement.fwr2 = self._extract_region(offsets.fwr2_start, offsets.fwr2_end)
        self.rearrangement.cdr2 = self._extract_region(offsets.cdr2_start, offsets.cdr2_end)
        self.rearrangement.fwr3 = self._extract_region(offsets.fwr3_start, offsets.fwr3_end)
        self.rearrangement.cdr3 = self._extract_region(offsets.cdr3_start, offsets.cdr3_end)
        self.rearrangement.fwr4 = self._extract_region(offsets.fwr4_start, offsets.fwr4_end)

        return self

    def with_aa_region_offsets(self, offsets: RegionOffsetsAA):
        self.aa_offsets = offsets

        self.rearrangement.fwr1_aa = self._extract_region_aa(offsets.fwr1_start_aa, offsets.fwr1_end_aa)
        self.rearrangement.cdr1_aa = self._extract_region_aa(offsets.cdr1_start_aa, offsets.cdr1_end_aa)
        self.rearrangement.fwr2_aa = self._extract_region_aa(offsets.fwr2_start_aa, offsets.fwr2_end_aa)
        self.rearrangement.cdr2_aa = self._extract_region_aa(offsets.cdr2_start_aa, offsets.cdr2_end_aa)
        self.rearrangement.fwr3_aa = self._extract_region_aa(offsets.fwr3_start_aa, offsets.fwr3_end_aa)
        self.rearrangement.cdr3_aa = self._extract_region_aa(offsets.cdr3_start_aa, offsets.cdr3_end_aa)
        self.rearrangement.fwr4_aa = self._extract_region_aa(offsets.fwr4_start_aa, offsets.fwr4_end_aa)

        return self

    def with_aa_scheme_alignment(self, alignment: SchemeAlignment):
        self.rearrangement.sequence_aa_scheme_cigar = fold_cigar(
            AlignmentString(alignment.alignment_str.replace("S", "").replace("N", ""))
        )
        return self

    def with_scheme_residue_mapping(self, mapping: dict[str, str]):
        self.rearrangement.scheme_residue_mapping = mapping
        return self

    def with_positional_scheme_mapping(self, mapping: dict[int, str]):
        self.rearrangement.positional_scheme_mapping = mapping
        return self

    def with_scheme_alignment_exc(self, exc: str):
        self.rearrangement.exc = exc
        return self

    def _extract_region(self, start: int, end: int):
        if start is None or end is None or start == -1 or end == -1:
            return None

        assert self.v_gene_alignment is not None
        start = start - self.v_gene_alignment.q_start
        end = end - self.v_gene_alignment.q_start

        if self.rearrangement.sequence_alignment:
            # subtract 1 because of 1 based indexing
            return self.rearrangement.sequence_alignment[start - 1 : end]
        return None

    def _extract_region_aa(self, start: int, end: int):
        if start is None or end is None or start == -1 or end == -1:
            return None

        if self.rearrangement.sequence_alignment_aa:
            return self.rearrangement.sequence_alignment_aa[start - 1 : end]
        return None

    def build(self):  # pylint: disable=too-many-statements
        if self.v_gene_alignment is None:
            return self.rearrangement

        alignment_end = self.j_gene_alignment.q_end if self.j_gene_alignment else self.v_gene_alignment.q_end
        self.rearrangement.sequence_alignment = self.sequence[self.v_gene_alignment.q_start : alignment_end]

        aln_len = len(self.rearrangement.sequence_alignment)
        v_germline_alignment = self.rearrangement.v_germline_alignment
        j_germline_alignment = self.rearrangement.j_germline_alignment or ""

        self.rearrangement.germline_alignment = (
            v_germline_alignment
            + ("N" * (aln_len - len(v_germline_alignment) - len(j_germline_alignment)))
            + j_germline_alignment
        )

        if self.rearrangement.sequence_alignment_aa:
            alignment_offset = self.v_gene_alignment.q_start + self.rearrangement.v_frame
            v_start_aa = 0
            v_end_aa = int((self.v_gene_alignment.q_end - alignment_offset) / 3)

            self.rearrangement.v_sequence_alignment_aa = self.rearrangement.sequence_alignment_aa[v_start_aa:v_end_aa]

            if self.j_gene_alignment is not None:
                j_start_aa = ceil((self.j_gene_alignment.q_start - alignment_offset) / 3)
                j_end_aa = int(self.j_gene_alignment.q_end - alignment_offset / 3)

                self.rearrangement.j_sequence_alignment_aa = self.rearrangement.sequence_alignment_aa[
                    j_start_aa:j_end_aa
                ]

            if self.rearrangement.cdr3 is not None:
                self.rearrangement.junction = self._extract_region(
                    self.rearrangement.cdr3_start - 3, self.rearrangement.cdr3_end + 3
                )
                self.rearrangement.junction_length = len(self.rearrangement.junction or "")

            if self.rearrangement.cdr3_aa is not None:
                jun_start_aa = self.aa_offsets.cdr3_start_aa - 1
                jun_end_aa = self.aa_offsets.cdr3_end_aa + 1

                self.rearrangement.junction_aa = self._extract_region_aa(jun_start_aa, jun_end_aa)
                self.rearrangement.junction_aa_length = len(self.rearrangement.junction_aa or "")

            # annotate
            self.rearrangement.stop_codon = "*" in self.rearrangement.sequence_alignment_aa
        self.rearrangement.complete_vdj = (
            self.rearrangement.fwr1_start
            and self.rearrangement.fwr1_start != -1
            and self.rearrangement.fwr4_end
            and self.rearrangement.fwr4_end != -1
        )

        # airr defines frameshift as sum of all insertions and deletions on gene alignment
        self.rearrangement.v_frameshift = self.rearrangement.v_cigar and has_frameshift(self.rearrangement.v_cigar)
        self.rearrangement.j_frameshift = self.rearrangement.j_cigar and has_frameshift(self.rearrangement.j_cigar)

        if self.rearrangement.j_frame is not None:
            j_frame_start = self.rearrangement.j_alignment_start - 1 + self.rearrangement.j_frame
            self.rearrangement.vj_in_frame = (j_frame_start % 3) == self.rearrangement.v_frame

            productive = not self.rearrangement.stop_codon and self.rearrangement.vj_in_frame
            self.rearrangement.productive = productive
        else:
            self.rearrangement.vj_in_frame = None
            self.rearrangement.productive = False

        self.rearrangement.additional_validation_flags = validate_airr_entry(self.rearrangement, scheme=self.scheme)

        return self.rearrangement



from riot_na.data.model import AlignmentString
from riot_na.schemes.collapse_alignment import (
    _collapse_ins_del_ordered,
    collapse_alignment_str,
    collapse_ins_del,
)


def test_collapse_alignment_str():
    # given
    alignment_str = AlignmentString("MDDIM")
    # when
    result = collapse_alignment_str(alignment_str)

    # then
    assert result == "MMDM"

    # given
    alignment_str = AlignmentString("MDDDIIIIIIM")
    # when
    result = collapse_alignment_str(alignment_str)

    # then
    assert result == "MMMIIIMM"

    # given
    alignment_str = AlignmentString("MDDDMIDIDIIM")
    # when
    result = collapse_alignment_str(alignment_str)

    # then
    assert result == "MDDDMMIIMM"

    # indels in the middle
    assert collapse_alignment_str(AlignmentString("MMMIIIIDDDMMM")) == "MMMMMIMMMM"

    # indels ordered
    assert collapse_alignment_str(AlignmentString("MMMIIIIDDDMMM"), ordered=True) == "MMMIMMMMMM"


def test_collapse_ins_del_ordered():
    assert _collapse_ins_del_ordered(list("DDDIIII")) == "MMMI"
    assert _collapse_ins_del_ordered(list("IIIDDDD")) == "MMMD"

    assert _collapse_ins_del_ordered(list("DDDDIII")) == "DMMM"
    assert _collapse_ins_del_ordered(list("IIIIDDD")) == "IMMM"

    assert _collapse_ins_del_ordered(list("IDDDI")) == "MDM"
    assert _collapse_ins_del_ordered(list("DIIID")) == "MIM"

    assert _collapse_ins_del_ordered(list("DDDIII")) == "MMM"
    assert _collapse_ins_del_ordered(list("IIIDDD")) == "MMM"

    assert _collapse_ins_del_ordered(list("IDIDID")) == "MMM"

    assert _collapse_ins_del_ordered(list("IDIIDIDDI")) == "MIMMM"

    assert _collapse_ins_del_ordered(list("IIIIDDDIIDIDDDIID")) == "IMMMIMMMMD"

    assert _collapse_ins_del_ordered(list("DDIDIDID")) == "DMMMD"

    assert _collapse_ins_del_ordered(list("IIDIDIDI")) == "IMMMI"


def test_collapse_insdel_insertions_in_junctions():
    assert collapse_ins_del(list("IDDDDDDDD ")) == "MDDDDDDD"

    assert collapse_ins_del(list("DIDDDDDDD ")) == "MDDDDDDD"

    assert collapse_ins_del(list("DIDDDDDID ")) == "MDDDDDM"

    assert collapse_ins_del(list("IDDDDDDID ")) == "MDDDDDM"

    assert collapse_ins_del(list("IDDDDDDDI ")) == "MDDDDDM"

    assert _collapse_ins_del_ordered(list("DIDDDDDID")) == "MDDDDMD"

    assert _collapse_ins_del_ordered(list("IMDDDDDDD")) == "MMDDDDDD"
    assert _collapse_ins_del_ordered(list("IMDDDDDMI")) == "MMDDDMM"


if __name__ == "__main__":
    test_collapse_alignment_str()
    test_collapse_ins_del_ordered()
    test_collapse_insdel_insertions_in_junctions()



# pylint: skip-file
# flake8: noqa
# isort: skip_file
__version__ = "1.1.0"
from riot_na.riot_na import Prefiltering
from riot_na.api.riot_numbering import (
    RiotNumberingAA,
    RiotNumberingNT,
    create_riot_aa,
    create_riot_nt,
)
from riot_na.api.api_mp import run_on_file_mp
from riot_na.data.scheme_regions import get_regions_definitions
from riot_na.data.scheme_definitions import Regions, ChainRegions
from riot_na.data.model import (
    AirrRearrangementEntryNT,
    AirrRearrangementEntryAA,
    ShortRegion,
    ChainType,
    Scheme,
    Organism,
)
from riot_na.api import utils



from itertools import groupby
from math import ceil
from typing import Sequence

from riot_na.data.model import AlignmentString


def collapse_ins_del(ops: Sequence[str]) -> AlignmentString:
    match_count = ops.count("M")
    deletions = ops.count("D")
    insertions = ops.count("I")
    extra_matches = min(deletions, insertions)
    match_count = match_count + extra_matches
    left_match_count = ceil(match_count / 2)
    right_match_count = match_count - left_match_count
    if deletions > insertions:
        return AlignmentString(f"{left_match_count*'M'}{(deletions-insertions) *'D'}{right_match_count*'M'}")
    if deletions < insertions:
        return AlignmentString(f"{left_match_count*'M'}{(insertions-deletions) *'I'}{right_match_count*'M'}")
    return AlignmentString(f"{match_count*'M'}")


def _collapse_ins_del_ordered(ops: Sequence[str]) -> AlignmentString:
    deletions_indices: list[int] = []
    insertion_indices: list[int] = []
    deletion_indices_to_change_to_match = set()
    insertion_indices_to_remove = set()
    last_op = ""

    for i, op in enumerate(ops):
        match op:
            case "D":
                if last_op == "I":
                    deletions_indices = []
                if insertion_indices:
                    deletion_indices_to_change_to_match.add(i)
                    insertion_indices_to_remove.add(insertion_indices.pop())
                else:
                    deletions_indices.append(i)
                last_op = op
            case "I":
                if last_op == "D":
                    insertion_indices = []
                if deletions_indices:
                    deletion_indices_to_change_to_match.add(deletions_indices.pop())
                    insertion_indices_to_remove.add(i)
                else:
                    insertion_indices.append(i)
                last_op = op

    ops_without_extra_insertions = [
        op if i not in deletion_indices_to_change_to_match else "M"
        for i, op in enumerate(ops)
        if i not in insertion_indices_to_remove
    ]
    return AlignmentString("".join(ops_without_extra_insertions))


def collapse_alignment_str(alignment_str: AlignmentString, ordered: bool = False) -> AlignmentString:
    collapse_ins_del_fn = _collapse_ins_del_ordered if ordered else collapse_ins_del
    res = []
    for key, group in groupby(alignment_str, lambda x: x == "M"):
        ops = list(group)
        if key:
            res.extend(ops)
        else:
            res.append(collapse_ins_del_fn(ops))
    return AlignmentString("".join(res))



from typing import Optional

from riot_na.data.model import (
    Locus,
    RegionOffsetsAA,
    RegionOffsetsNT,
    Scheme,
    SchemeAlignment,
    ShortRegion,
)
from riot_na.data.scheme_regions import get_region


def infer_region_offsets(
    scheme_alignment: SchemeAlignment, scheme: Scheme, locus: Locus
) -> dict[ShortRegion, list[int]]:
    result: dict[ShortRegion, list[int]] = {}

    scheme_position = 0
    query_position = scheme_alignment.q_start or 0
    insertion_counter = 0
    alignment_str = scheme_alignment.alignment_str

    region_id = ShortRegion.FW1

    for op in alignment_str:
        if op == "M":
            insertion_counter = 0
            query_position = query_position + 1
            scheme_position = scheme_position + 1
            region_id = get_region(scheme_position, scheme, locus)

            residues = result.get(region_id, [])
            residues.append(query_position)
            result[region_id] = residues

        elif op in {"D", "N"}:
            insertion_counter = 0
            scheme_position = scheme_position + 1
            region_id = get_region(scheme_position, scheme, locus)
        elif op == "I":
            insertion_counter = insertion_counter + 1
            query_position = query_position + 1
            residues = result.get(region_id, [])
            residues.append(query_position)
            result[region_id] = residues

    return result


def get_first_aa(region_positions: Optional[list[int]]) -> int:
    if not region_positions or len(region_positions) == 0:
        return -1

    return region_positions[0]


def get_last_aa(region_positions: Optional[list[int]]) -> int:
    if not region_positions or len(region_positions) == 0:
        return -1

    return region_positions[-1]


def infer_aa_region_offsets(aa_regions: dict[ShortRegion, list[int]]) -> RegionOffsetsAA:
    return RegionOffsetsAA(
        fwr1_start_aa=get_first_aa(aa_regions.get(ShortRegion.FW1)),
        fwr1_end_aa=get_last_aa(aa_regions.get(ShortRegion.FW1)),
        cdr1_start_aa=get_first_aa(aa_regions.get(ShortRegion.CDR1)),
        cdr1_end_aa=get_last_aa(aa_regions.get(ShortRegion.CDR1)),
        fwr2_start_aa=get_first_aa(aa_regions.get(ShortRegion.FW2)),
        fwr2_end_aa=get_last_aa(aa_regions.get(ShortRegion.FW2)),
        cdr2_start_aa=get_first_aa(aa_regions.get(ShortRegion.CDR2)),
        cdr2_end_aa=get_last_aa(aa_regions.get(ShortRegion.CDR2)),
        fwr3_start_aa=get_first_aa(aa_regions.get(ShortRegion.FW3)),
        fwr3_end_aa=get_last_aa(aa_regions.get(ShortRegion.FW3)),
        cdr3_start_aa=get_first_aa(aa_regions.get(ShortRegion.CDR3)),
        cdr3_end_aa=get_last_aa(aa_regions.get(ShortRegion.CDR3)),
        fwr4_start_aa=get_first_aa(aa_regions.get(ShortRegion.FW4)),
        fwr4_end_aa=get_last_aa(aa_regions.get(ShortRegion.FW4)),
    )


def get_first_nt(region: Optional[list[int]], offset: int) -> int:
    if region is None:
        return -1

    return region[0] * 3 - 2 + offset


def get_last_nt(region: Optional[list[int]], offset: int) -> int:
    if region is None:
        return -1

    return region[-1] * 3 + offset


def infer_nt_region_offsets(
    aa_regions: dict[ShortRegion, list[int]], nt_alignment_start: int = 0, reading_frame: int = 0
) -> RegionOffsetsNT:
    # nt     1 2 3 4 5 6 7
    # pt       1     2
    # start  |
    # end              |
    offset = nt_alignment_start + reading_frame

    return RegionOffsetsNT(
        fwr1_start=get_first_nt(aa_regions.get(ShortRegion.FW1), offset),
        fwr1_end=get_last_nt(aa_regions.get(ShortRegion.FW1), offset),
        cdr1_start=get_first_nt(aa_regions.get(ShortRegion.CDR1), offset),
        cdr1_end=get_last_nt(aa_regions.get(ShortRegion.CDR1), offset),
        fwr2_start=get_first_nt(aa_regions.get(ShortRegion.FW2), offset),
        fwr2_end=get_last_nt(aa_regions.get(ShortRegion.FW2), offset),
        cdr2_start=get_first_nt(aa_regions.get(ShortRegion.CDR2), offset),
        cdr2_end=get_last_nt(aa_regions.get(ShortRegion.CDR2), offset),
        fwr3_start=get_first_nt(aa_regions.get(ShortRegion.FW3), offset),
        fwr3_end=get_last_nt(aa_regions.get(ShortRegion.FW3), offset),
        cdr3_start=get_first_nt(aa_regions.get(ShortRegion.CDR3), offset),
        cdr3_end=get_last_nt(aa_regions.get(ShortRegion.CDR3), offset),
        fwr4_start=get_first_nt(aa_regions.get(ShortRegion.FW4), offset),
        fwr4_end=get_last_nt(aa_regions.get(ShortRegion.FW4), offset),
    )



from riot_na import Prefiltering
from riot_na.data.model import (
    Gene,
    Locus,
    Organism,
    SpeciesGeneMatch,
    SpeciesPrefilteringResult,
)


class MultiSpeciesPrefiltering:
    def __init__(self, all_genes: list[Gene], kmer_size: int, distance_threshold: int, top_n: int, modulo_n: int):
        self.all_genes = dict(
            map(lambda gene: (gene.species + "|" + gene.locus + "|" + gene.name, gene.sequence), all_genes)
        )

        self.prefiltering = Prefiltering(
            self.all_genes,
            top_n=top_n,
            kmer_size=kmer_size,
            distance_threshold=distance_threshold,
            modulo_n=modulo_n,
        )

    def calculate_top_matches_with_rev_comp(self, query: str) -> SpeciesPrefilteringResult:
        mapped_matches = []
        try:
            raw_result = self.prefiltering.calculate_top_matches_with_rev_comp(query)
        except ValueError:
            return SpeciesPrefilteringResult(query=query, rev_comp_query="", top_matches=[])

        for match in raw_result.top_matches:
            (species, locus, name) = match.gene_id.split("|")
            mapped_matches.append(
                SpeciesGeneMatch(
                    gene_id=name,
                    rev_comp=match.rev_comp,
                    coverage=match.coverage,
                    species=Organism(species),
                    locus=Locus(locus),
                )
            )
        return SpeciesPrefilteringResult(
            query=query, rev_comp_query=raw_result.rev_comp_query, top_matches=mapped_matches
        )

    def calculate_top_matches(self, query: str) -> SpeciesPrefilteringResult:
        mapped_matches = []
        try:
            raw_result = self.prefiltering.calculate_top_matches(query)

        except ValueError:
            return SpeciesPrefilteringResult(query=query, rev_comp_query="", top_matches=[])

        for match in raw_result.top_matches:
            (species, locus, name) = match.gene_id.split("|")
            mapped_matches.append(
                SpeciesGeneMatch(
                    gene_id=name,
                    rev_comp=match.rev_comp,
                    coverage=match.coverage,
                    species=Organism(species),
                    locus=Locus(locus),
                )
            )

        return SpeciesPrefilteringResult(
            query=query, rev_comp_query=raw_result.rev_comp_query, top_matches=mapped_matches
        )



import json
from dataclasses import asdict
from pathlib import Path
from typing import Optional

from skbio.alignment import StripedSmithWaterman  # type: ignore

from riot_na.alignment.alignment_utils import infer_reading_frame, translate
from riot_na.alignment.gene_aligner import (
    AA_ALIGNER_PARAMS,
    GeneAlignerAA,
    create_aa_j_gene_aligner,
    create_aa_v_gene_aligner,
)
from riot_na.alignment.skbio_alignment import align_aa
from riot_na.config import GENE_DB_DIR
from riot_na.data.model import (
    AlignmentEntryAA,
    AlignmentsAA,
    AlignmentsNT,
    Locus,
    Organism,
    TranslatedAlignmentsNT,
)


class VJAlignerAA:
    def __init__(
        self,
        v_aligner: GeneAlignerAA,
        j_aligners: dict[Organism, dict[Locus, GeneAlignerAA]],
    ) -> None:
        self.v_aligner = v_aligner
        self.j_aligners = j_aligners

    def produce_aa_alignments(self, query: str) -> AlignmentsAA:
        result = AlignmentsAA()

        v_aa_alignment = self.v_aligner.align(query)

        if v_aa_alignment is None:
            return result

        result.v = v_aa_alignment

        # mask alignments
        query_j = query[v_aa_alignment.q_end :]

        # align j genes
        species = v_aa_alignment.species
        locus = v_aa_alignment.locus

        j_aligner = self.j_aligners[species][locus]
        j_aa_alignment = j_aligner.align(query_j)

        if j_aa_alignment is None:
            return result

        assert j_aa_alignment is not None
        result.j = j_aa_alignment

        return result


class VJAlignmentTranslatorAA:
    def __init__(
        self,
        v_aligner: GeneAlignerAA,
        j_aligners: dict[Organism, dict[Locus, GeneAlignerAA]],
    ) -> None:
        self.v_aligner = v_aligner
        self.j_aligners = j_aligners

    def translate_nt_alignments(self, alignments: AlignmentsNT) -> Optional[TranslatedAlignmentsNT]:
        if alignments.v:
            # Query might have been reverse complemented during V alignment.
            query_sequence = alignments.v.q_seq
            v_alignment = alignments.v
            assert v_alignment.reading_frame is not None
            reading_frame = infer_reading_frame(t_start=v_alignment.t_start, t_frame=v_alignment.reading_frame)

            # translate sequence to aa
            alignment_start = alignments.v.q_start
            alignment_end = alignments.j.q_end if alignments.j else alignments.v.q_end

            aligned_sequence = query_sequence[alignment_start:alignment_end]
            aligned_sequence_aa = translate(aligned_sequence, reading_frame)

            species = v_alignment.species
            v_gene = self.v_aligner.gene_lookup[species + "|" + v_alignment.target_id]

            locus = v_alignment.locus
            v_gene_aa = v_gene.sequence

            j_gene_id = alignments.j.target_id if alignments.j else None
            j_gene = self.j_aligners[species][locus].gene_lookup[species + "|" + j_gene_id] if j_gene_id else None

            j_gene_aa = j_gene.sequence if j_gene else None

            v_aligner = StripedSmithWaterman(aligned_sequence_aa, **AA_ALIGNER_PARAMS)
            v_aa_alignment = align_aa(
                v_aligner, aligned_sequence_aa, alignments.v.target_id, v_gene_aa, calculate_score=False
            )
            assert (
                v_aa_alignment.t_start != -1
                and v_aa_alignment.t_end != -1
                and v_aa_alignment.q_start != -1
                and v_aa_alignment.q_end != -1
            ), "Query translation failed: could not align sequence_aa to germline_aa"

            masked_aa = aligned_sequence_aa[v_aa_alignment.q_end :]
            j_aligner = StripedSmithWaterman(masked_aa, **AA_ALIGNER_PARAMS)
            j_aa_alignment = (
                align_aa(j_aligner, masked_aa, j_gene_id, j_gene_aa, calculate_score=False)
                if j_gene_id and j_gene_aa
                else None
            )

            extended_v_aa_alignment = AlignmentEntryAA(
                target_id=v_alignment.target_id,
                alignment_score=v_aa_alignment.alignment_score,
                seq_identity=v_aa_alignment.seq_identity,
                e_value=v_aa_alignment.e_value,
                q_start=v_aa_alignment.q_start,
                q_end=v_aa_alignment.q_end,
                t_start=v_aa_alignment.t_start,
                t_end=v_aa_alignment.t_end,
                cigar=v_aa_alignment.cigar,
                species=v_alignment.species,
                locus=v_alignment.locus,
                q_seq=aligned_sequence_aa,
                t_seq=v_gene_aa,
            )

            if j_aa_alignment:
                assert j_gene_aa
                assert j_gene_id

                extended_j_aa_alignment = AlignmentEntryAA(
                    target_id=j_gene_id,
                    alignment_score=j_aa_alignment.alignment_score,
                    seq_identity=j_aa_alignment.seq_identity,
                    e_value=j_aa_alignment.e_value,
                    q_start=j_aa_alignment.q_start,
                    q_end=j_aa_alignment.q_end,
                    t_start=j_aa_alignment.t_start,
                    t_end=j_aa_alignment.t_end,
                    cigar=j_aa_alignment.cigar,
                    species=v_alignment.species,
                    locus=v_alignment.locus,
                    q_seq=masked_aa,
                    t_seq=j_gene_aa,
                )
            else:
                extended_j_aa_alignment = None

            aa_alignments = AlignmentsAA(v=extended_v_aa_alignment, j=extended_j_aa_alignment)

            return TranslatedAlignmentsNT(
                translated_query=aligned_sequence_aa, reading_frame=reading_frame, aa_alignments=aa_alignments
            )

        return None


def create_vj_aligner_aa(allowed_species: Optional[list[Organism]] = None, db_dir: Path = GENE_DB_DIR) -> VJAlignerAA:
    if not allowed_species:
        allowed_species = list(Organism)

    aa_genes_dir = db_dir / "gene_db" / "aa_genes_deduplicated"
    v_aligner = create_aa_v_gene_aligner(allowed_species=allowed_species, aa_genes_dir=aa_genes_dir)

    j_aligners: dict[Organism, dict[Locus, GeneAlignerAA]] = {}

    for organism in allowed_species:
        for locus in Locus:
            j_aligner = create_aa_j_gene_aligner(organism=organism, locus=locus, aa_genes_dir=aa_genes_dir)
            organism_aligners = j_aligners.get(organism, {})
            organism_aligners[locus] = j_aligner
            j_aligners[organism] = organism_aligners

    vj_aligner = VJAlignerAA(v_aligner, j_aligners)

    return vj_aligner


def create_vj_alignment_translator_aa(
    allowed_species: Optional[list[Organism]] = None, db_dir: Path = GENE_DB_DIR
) -> VJAlignmentTranslatorAA:
    if not allowed_species:
        allowed_species = list(Organism)

    aa_genes_dir = db_dir / "gene_db" / "aa_genes"
    v_aligner = create_aa_v_gene_aligner(allowed_species=allowed_species, aa_genes_dir=aa_genes_dir)

    j_aligners: dict[Organism, dict[Locus, GeneAlignerAA]] = {}

    for organism in allowed_species:
        for locus in Locus:
            j_aligner = create_aa_j_gene_aligner(organism=organism, locus=locus, aa_genes_dir=aa_genes_dir)
            organism_aligners = j_aligners.get(organism, {})
            organism_aligners[locus] = j_aligner
            j_aligners[organism] = organism_aligners

    vj_aligner = VJAlignmentTranslatorAA(v_aligner, j_aligners)

    return vj_aligner


if __name__ == "__main__":
    # case 1 detect organism

    # given query
    SAMPLE_QUERY = (
        "LRRPCPSPALSLMAPSVVTTGTGSGSPQGRDWSGLGIYTIVGAPTTTPPSRVESPYQTPKNHFSLKLRSVTAVDTAVYYCARWGHFDTSGYFVVDYWGQGTLVTVSS"
    )

    vdj_alnr = create_vj_aligner_aa()

    nt_alignments = vdj_alnr.produce_aa_alignments(SAMPLE_QUERY)
    print(json.dumps(asdict(nt_alignments), indent=4))



from functools import cache
from typing import Final, TypedDict

from riot_na.data.model import ChainType, Scheme, ShortRegion

# Number of scheme positions for each chain type
IMGT_POSITIONS: Final[int] = 128
KABAT_POSITIONS_LIGHT: Final[int] = 107
KABAT_POSITIONS_HEAVY: Final[int] = 113
CHOTHIA_POSITIONS_LIGHT: Final[int] = 107
CHOTHIA_POSITIONS_HEAVY: Final[int] = 113
MARTIN_POSITIONS_LIGHT: Final[int] = 107
MARTIN_POSITIONS_HEAVY: Final[int] = 113

# CDR insertions / deletions positions
IMGT_INDEL_POSITIONS: Final[list[int]] = [32, 60, 111]
KABAT_INDEL_POSITIONS_HEAVY: Final[list[int]] = [35, 52, 100]
KABAT_INDEL_POSITIONS_LIGHT: Final[list[int]] = [27, 52, 95]
CHOTHIA_INDEL_POSITIONS_HEAVY: Final[list[int]] = [31, 52, 100]
CHOTHIA_INDEL_POSITIONS_LIGHT: Final[list[int]] = [30, 52, 95]
MARTIN_INDEL_POSITIONS_HEAVY: Final[list[int]] = [31, 52, 100]
MARTIN_INDEL_POSITIONS_LIGHT: Final[list[int]] = [30, 52, 95]


# Region position ranges
class MinMaxRange(TypedDict):
    min: int
    max: int


ChainRegions = dict[ShortRegion, MinMaxRange]
Regions = dict[ChainType, ChainRegions]

IMGT_REGIONS: Final[Regions] = {
    ChainType.HEAVY: {
        ShortRegion.FW1: {"min": 1, "max": 26},
        ShortRegion.CDR1: {"min": 27, "max": 38},
        ShortRegion.FW2: {"min": 39, "max": 55},
        ShortRegion.CDR2: {"min": 56, "max": 65},
        ShortRegion.FW3: {"min": 66, "max": 104},
        ShortRegion.CDR3: {"min": 105, "max": 117},
        ShortRegion.FW4: {"min": 118, "max": 128},
    },
    ChainType.LIGHT: {
        ShortRegion.FW1: {"min": 1, "max": 26},
        ShortRegion.CDR1: {"min": 27, "max": 38},
        ShortRegion.FW2: {"min": 39, "max": 55},
        ShortRegion.CDR2: {"min": 56, "max": 65},
        ShortRegion.FW3: {"min": 66, "max": 104},
        ShortRegion.CDR3: {"min": 105, "max": 117},
        ShortRegion.FW4: {"min": 118, "max": 128},
    },
}

KABAT_REGIONS: Final[Regions] = {
    ChainType.HEAVY: {
        ShortRegion.FW1: {"min": 1, "max": 30},
        ShortRegion.CDR1: {"min": 31, "max": 35},
        ShortRegion.FW2: {"min": 36, "max": 49},
        ShortRegion.CDR2: {"min": 50, "max": 65},
        ShortRegion.FW3: {"min": 66, "max": 94},
        ShortRegion.CDR3: {"min": 95, "max": 102},
        ShortRegion.FW4: {"min": 103, "max": 113},
    },
    ChainType.LIGHT: {
        ShortRegion.FW1: {"min": 1, "max": 23},
        ShortRegion.CDR1: {"min": 24, "max": 34},
        ShortRegion.FW2: {"min": 35, "max": 49},
        ShortRegion.CDR2: {"min": 50, "max": 56},
        ShortRegion.FW3: {"min": 57, "max": 88},
        ShortRegion.CDR3: {"min": 89, "max": 97},
        ShortRegion.FW4: {"min": 98, "max": 107},
    },
}


CHOTHIA_REGIONS: Final[Regions] = {
    ChainType.HEAVY: {
        ShortRegion.FW1: {"min": 1, "max": 25},
        ShortRegion.CDR1: {"min": 26, "max": 32},
        ShortRegion.FW2: {"min": 33, "max": 51},
        ShortRegion.CDR2: {"min": 52, "max": 56},
        ShortRegion.FW3: {"min": 57, "max": 95},
        ShortRegion.CDR3: {"min": 96, "max": 101},
        ShortRegion.FW4: {"min": 102, "max": 113},
    },
    ChainType.LIGHT: {
        ShortRegion.FW1: {"min": 1, "max": 25},
        ShortRegion.CDR1: {"min": 26, "max": 32},
        ShortRegion.FW2: {"min": 33, "max": 49},
        ShortRegion.CDR2: {"min": 50, "max": 52},
        ShortRegion.FW3: {"min": 53, "max": 90},
        ShortRegion.CDR3: {"min": 91, "max": 96},
        ShortRegion.FW4: {"min": 97, "max": 107},
    },
}

MARTIN_REGIONS: Final[Regions] = {
    ChainType.HEAVY: {
        ShortRegion.FW1: {"min": 1, "max": 25},
        ShortRegion.CDR1: {"min": 26, "max": 32},
        ShortRegion.FW2: {"min": 33, "max": 51},
        ShortRegion.CDR2: {"min": 52, "max": 56},
        ShortRegion.FW3: {"min": 57, "max": 95},
        ShortRegion.CDR3: {"min": 96, "max": 101},
        ShortRegion.FW4: {"min": 102, "max": 113},
    },
    ChainType.LIGHT: {
        ShortRegion.FW1: {"min": 1, "max": 25},
        ShortRegion.CDR1: {"min": 26, "max": 32},
        ShortRegion.FW2: {"min": 33, "max": 49},
        ShortRegion.CDR2: {"min": 50, "max": 52},
        ShortRegion.FW3: {"min": 53, "max": 90},
        ShortRegion.CDR3: {"min": 91, "max": 96},
        ShortRegion.FW4: {"min": 97, "max": 107},
    },
}


@cache
def get_legal_positions(chain_type: ChainType, scheme: Scheme) -> int:
    match scheme, chain_type:
        case Scheme.IMGT, ChainType.HEAVY:
            return IMGT_POSITIONS
        case Scheme.IMGT, ChainType.LIGHT:
            return IMGT_POSITIONS
        case Scheme.KABAT, ChainType.HEAVY:
            return KABAT_POSITIONS_HEAVY
        case Scheme.KABAT, ChainType.LIGHT:
            return KABAT_POSITIONS_LIGHT
        case Scheme.CHOTHIA, ChainType.HEAVY:
            return CHOTHIA_POSITIONS_HEAVY
        case Scheme.CHOTHIA, ChainType.LIGHT:
            return CHOTHIA_POSITIONS_LIGHT
        case Scheme.MARTIN, ChainType.HEAVY:
            return MARTIN_POSITIONS_HEAVY
        case Scheme.MARTIN, ChainType.LIGHT:
            return MARTIN_POSITIONS_LIGHT
        case _:
            raise ValueError(f"Unknown scheme {scheme} or chain type {chain_type}")


@cache
def get_indels_positions(scheme: Scheme, chain_type: ChainType) -> list[int]:
    # this is scheme specific
    match (scheme, chain_type):
        case Scheme.IMGT, _:
            return IMGT_INDEL_POSITIONS
        case Scheme.KABAT, ChainType.HEAVY:
            return KABAT_INDEL_POSITIONS_HEAVY
        case Scheme.KABAT, ChainType.LIGHT:
            return KABAT_INDEL_POSITIONS_LIGHT
        case Scheme.CHOTHIA, ChainType.HEAVY:
            return CHOTHIA_INDEL_POSITIONS_HEAVY
        case Scheme.CHOTHIA, ChainType.LIGHT:
            return CHOTHIA_INDEL_POSITIONS_LIGHT
        case Scheme.MARTIN, ChainType.HEAVY:
            return MARTIN_INDEL_POSITIONS_HEAVY
        case Scheme.MARTIN, ChainType.LIGHT:
            return MARTIN_INDEL_POSITIONS_LIGHT

    raise ValueError(f"Unknown scheme {scheme}")



# type: ignore
import json
import multiprocessing as mp
import re
from functools import partial

import pandas as pd
from tqdm import tqdm

from riot_na.alignment.alignment_utils import get_cigar_op_groups

GRID_SEARCH_PARAMS = ["top", "kmer_size", "distance_threshold", "modulo", "gap_open", "gap_ext", "x_drop"]


def parse_file_name(path, params=None):
    if params is None:
        params = GRID_SEARCH_PARAMS
    regex = "_".join([f"{param}_([0-9]*)" for param in params])
    match = re.search(regex, path.stem)
    return match.groups()


def _extract_gene_match(matches_string, top_n, with_rev_comp):
    matches = json.loads(matches_string)[:top_n]
    if with_rev_comp:
        return {(match["gene_id"], match["rev_comp"]) for match in matches}
    return {match["gene_id"] for match in matches}


def _agg_grid_search_worker(path, params, top_n, ground_truth_df, with_rev_comp):
    df = pd.read_csv(path, index_col=0, usecols=["sequence_id", "best_genes"])
    df["best_genes"] = df["best_genes"].apply(partial(_extract_gene_match, top_n=top_n, with_rev_comp=with_rev_comp))
    df = df.join(ground_truth_df, rsuffix="_true", how="inner")
    df["is_match"] = df.apply(lambda s: s["target_id"] in s["best_genes"], axis=1)
    match_percent = df["is_match"].value_counts(normalize=True).loc[True]
    return {**dict(zip(params, parse_file_name(path, params))), "match_percent": match_percent}


def aggregate_grid_search_results(results_dir, params, top_n, ground_truth_df, with_rev_comp):
    worker_partial = partial(
        _agg_grid_search_worker,
        params=params,
        top_n=top_n,
        ground_truth_df=ground_truth_df,
        with_rev_comp=with_rev_comp,
    )
    paths = list(results_dir.glob("top_*"))
    with mp.Pool(4) as pool:
        res = list(tqdm(pool.imap(worker_partial, paths), total=len(paths)))
        res_df = pd.DataFrame.from_records(res)
        res_df["top"] = top_n
        res_df[params] = res_df[params].astype(int)
        return res_df


def format_cigar(cigar: str, query: str, target: str):
    cigar_items = get_cigar_op_groups(cigar)
    out = ""
    pos = 0
    for cnt, operation in cigar_items:
        if operation == "M":
            for i in range(cnt):
                if query[pos + i] == target[pos + i]:
                    out += "|"
                else:
                    out += "X"
        elif operation == "=":
            out += cnt * "|"
        elif operation == "X":
            out += cnt * "X"
        elif operation == "I":
            target = target[:pos] + cnt * "-" + target[pos:]
            out += cnt * "+"
        elif operation == "D":
            query = query[:pos] + cnt * "-" + query[pos:]
            out += cnt * "-"
        else:
            out += cnt * "?"
        pos += cnt
    return out, query, target


def calculate_cigar_op_sum(cigar: str) -> int:
    return sum(len for len, _ in get_cigar_op_groups(cigar))


def display_riot_alignment(row):
    print(f"ID: {row.name}")
    print(f"rev_comp: {row['rev_comp']}")
    print()

    t_start = row["t_start"]
    t_end = row["t_end"]
    q_start = row["q_start"]
    q_end = row["q_end"]
    seq_identity = row["seq_identity"]
    bit_score = row["bit_score"]
    e_value = row["e_value"]
    target = row["target"][t_start:t_end]
    query = row["sequence"][q_start:q_end]
    cigar = row["cigar"]
    cigar_op_sum = calculate_cigar_op_sum(cigar)

    print(f"Riot: {row['target_id']}")
    print(f"{t_start=}, {t_end=}, {len(target)=}, {q_start=}, {q_end=}, {len(query)=}, {cigar=}, {cigar_op_sum=}")
    print(f"{seq_identity=}")
    print(f"{bit_score=}")
    print(f"{e_value=}")
    cigar_exploded, query, target = format_cigar(cigar, query, target)

    print("Target: ", target)
    print("        ", cigar_exploded)
    print("Query:  ", query)



from typing import Optional

from skbio.alignment import StripedSmithWaterman  # type: ignore

from riot_na.alignment.alignment_metrics import (
    GumbellParams,
    calculate_seq_identity,
    compute_bit_score,
    compute_evalue,
)
from riot_na.data.model import Cigar, InternalAlignmentEntry, InternalAlignmentEntryAA


# Extracted aligner call for profiling purposes.
def _align(aligner, target):
    res = aligner(target)
    return res


def align(
    aligner: StripedSmithWaterman,
    target_id: str,
    target: str,
    db_length: int,
    query: str,
    rev_comp: bool = False,
) -> InternalAlignmentEntry:
    res = _align(aligner, target)
    bit_score = compute_bit_score(res["optimal_alignment_score"])
    e_value = compute_evalue(len(query), db_length, res["optimal_alignment_score"])
    # Need to add 1 because of: https://github.com/biocore/scikit-bio/issues/1340

    q_start = res["query_begin"]
    q_end = res["query_end"] + 1
    t_start = res["target_begin"]
    t_end = res["target_end_optimal"] + 1

    # seq_identity = calculate_seq_identity(res["cigar"], query[q_start:q_end], target[t_start:t_end])
    return InternalAlignmentEntry(
        target_id=target_id,
        alignment_score=bit_score,
        seq_identity=calculate_seq_identity(res["cigar"], query, target, q_start, t_start),
        e_value=e_value,
        q_start=q_start,
        q_end=q_end,
        q_len=q_end - q_start,
        t_start=t_start,
        t_end=t_end,
        t_len=t_end - t_start,
        cigar=res["cigar"],
        query=query,
        rev_comp=rev_comp,
    )


def align_aa(
    aligner: StripedSmithWaterman,
    query: str,
    target_id: str,
    target: str,
    db_length: Optional[int] = None,
    calculate_score: bool = True,
) -> InternalAlignmentEntryAA:
    res = aligner(target)

    q_start = res["query_begin"]
    q_end = res["query_end"] + 1
    t_start = res["target_begin"]
    t_end = res["target_end_optimal"] + 1

    if calculate_score:
        assert db_length is not None
        bit_score = compute_bit_score(res["optimal_alignment_score"], gumbell_params=GumbellParams.AA)
        e_value = compute_evalue(len(query), db_length, res["optimal_alignment_score"])
        seq_identity = calculate_seq_identity(res["cigar"], query, target, q_start, t_start) if query else 0
    else:
        bit_score = None
        e_value = None
        seq_identity = None

    return InternalAlignmentEntryAA(
        target_id=target_id,
        alignment_score=bit_score,
        seq_identity=seq_identity,
        e_value=e_value,
        q_start=q_start,
        q_end=q_end,
        t_start=t_start,
        t_end=t_end,
        cigar=Cigar(res["cigar"]),
    )



import json
import os
from pathlib import Path

from dotenv import load_dotenv

if bool(json.loads(os.getenv("LOAD_DOTENV", "true"))):
    # this is somewhat a hack - on production code is run directly from a .zip file. Because of that dotenv is not able
    # to find an .env file and will crash. Whole env is being set
    try:
        load_dotenv()
    except OSError:
        pass

GENE_DB_DIR = Path(os.getenv("GENE_DB_DIR", Path(__file__).parent / "databases"))



class GeneMatch:
    gene_id: str
    rev_comp: bool
    coverage: int

class PrefilteringResult:
    query: str
    rev_comp_query: str
    top_matches: list[GeneMatch]

class Prefiltering:
    def __init__(self, genes: dict[str, str], kmer_size: int, distance_threshold: int, top_n: int, modulo_n: int): ...
    def calculate_top_matches_with_rev_comp(self, query: str) -> PrefilteringResult: ...
    def calculate_top_matches(self, query: str) -> PrefilteringResult: ...



import json
from dataclasses import dataclass
from enum import Enum, unique
from typing import Optional, TypeVar

from riot_na.common.serialization_utils import base64_encode


class Cigar(str):
    def __new__(cls, *args, **kwargs):
        return str.__new__(cls, *args, **kwargs)


class AlignmentString(str):
    def __new__(cls, *args, **kwargs):
        return str.__new__(cls, *args, **kwargs)


@dataclass(frozen=True)
class InternalAlignmentEntry:  # pylint: disable=too-many-instance-attributes
    """https://github.com/soedinglab/MMseqs2/wiki#internal-alignment-format"""

    target_id: str
    alignment_score: float
    seq_identity: float
    e_value: float
    q_start: int
    q_end: int
    q_len: int
    t_start: int
    t_end: int
    t_len: int
    cigar: str
    query: str
    rev_comp: bool

    def __lt__(self, other):
        if self.e_value == other.e_value:
            return (self.t_end - self.t_start) < (other.t_end - other.t_start)
        return self.e_value < other.e_value

    def __gt__(self, other):
        if self.e_value == other.e_value:
            return (self.t_end - self.t_start) > (other.t_end - other.t_start)
        return self.e_value > other.e_value


@dataclass(frozen=True)
class InternalAlignmentEntryAA:  # pylint: disable=too-many-instance-attributes
    target_id: str
    alignment_score: Optional[float]
    seq_identity: Optional[float]
    e_value: Optional[float]
    q_start: int
    q_end: int
    t_start: int
    t_end: int
    cigar: Cigar

    def __lt__(self, other):
        if self.e_value == other.e_value:
            return (self.t_end - self.t_start) < (other.t_end - other.t_start)
        return self.e_value < other.e_value

    def __gt__(self, other):
        if self.e_value == other.e_value:
            return (self.t_end - self.t_start) > (other.t_end - other.t_start)
        return self.e_value > other.e_value


@dataclass(frozen=True)
class SchemeAlignment:
    target_id: str
    q_start: int
    q_end: int
    alignment_str: AlignmentString
    exc: Optional[str] = None


class Locus(str, Enum):
    IGH = "igh"
    IGL = "igl"
    IGK = "igk"


class ChainType(str, Enum):
    HEAVY = "H"
    LIGHT = "L"

    @classmethod
    def from_locus(cls, locus: Locus):
        match locus:
            case Locus.IGH:
                return cls.HEAVY
            case _:
                return cls.LIGHT


@unique
class Scheme(str, Enum):
    KABAT = "kabat"
    CHOTHIA = "chothia"
    IMGT = "imgt"
    MARTIN = "martin"


@unique
class Organism(str, Enum):
    HOMO_SAPIENS = "human"
    MUS_MUSCULUS = "mouse"


class ShortRegion(str, Enum):
    FW1 = "fw1"
    FW2 = "fw2"
    FW3 = "fw3"
    FW4 = "fw4"
    CDR1 = "cdr1"
    CDR2 = "cdr2"
    CDR3 = "cdr3"


EntryValidationResults = dict[str, Optional[bool]]


@dataclass
class AirrRearrangementEntryNT:  # pylint: disable=too-many-instance-attributes
    """https://docs.airr-community.org/en/stable/datarep/rearrangements.html"""

    sequence_header: str
    sequence: str
    numbering_scheme: str
    locus: Optional[str] = None
    stop_codon: Optional[bool] = None
    vj_in_frame: Optional[bool] = None
    v_frameshift: Optional[bool] = None
    j_frameshift: Optional[bool] = None
    productive: Optional[bool] = None
    rev_comp: Optional[bool] = None
    complete_vdj: Optional[bool] = None
    v_call: Optional[str] = None
    d_call: Optional[str] = None
    j_call: Optional[str] = None
    c_call: Optional[str] = None
    v_frame: Optional[int] = None
    j_frame: Optional[int] = None
    sequence_alignment: Optional[str] = None
    germline_alignment: Optional[str] = None
    sequence_alignment_aa: Optional[str] = None
    v_alignment_start: Optional[int] = None
    v_alignment_end: Optional[int] = None
    d_alignment_start: Optional[int] = None
    d_alignment_end: Optional[int] = None
    j_alignment_start: Optional[int] = None
    j_alignment_end: Optional[int] = None
    c_alignment_start: Optional[int] = None
    c_alignment_end: Optional[int] = None
    v_sequence_alignment: Optional[str] = None
    v_sequence_alignment_aa: Optional[str] = None
    v_germline_alignment: Optional[str] = None
    d_sequence_alignment: Optional[str] = None

    d_germline_alignment: Optional[str] = None
    j_sequence_alignment: Optional[str] = None
    j_sequence_alignment_aa: Optional[str] = None
    j_germline_alignment: Optional[str] = None
    c_sequence_alignment: Optional[str] = None
    c_germline_alignment: Optional[str] = None
    fwr1: Optional[str] = None
    fwr1_aa: Optional[str] = None
    cdr1: Optional[str] = None
    cdr1_aa: Optional[str] = None
    fwr2: Optional[str] = None
    fwr2_aa: Optional[str] = None
    cdr2: Optional[str] = None
    cdr2_aa: Optional[str] = None
    fwr3: Optional[str] = None
    fwr3_aa: Optional[str] = None
    cdr3: Optional[str] = None
    cdr3_aa: Optional[str] = None
    fwr4: Optional[str] = None
    fwr4_aa: Optional[str] = None
    junction: Optional[str] = None
    junction_aa: Optional[str] = None
    junction_length: Optional[int] = None
    junction_aa_length: Optional[int] = None
    v_score: Optional[float] = None
    d_score: Optional[float] = None
    j_score: Optional[float] = None
    c_score: Optional[float] = None

    v_cigar: Optional[str] = None
    d_cigar: Optional[str] = None
    j_cigar: Optional[str] = None
    c_cigar: Optional[str] = None

    v_support: Optional[float] = None
    d_support: Optional[float] = None
    j_support: Optional[float] = None
    c_support: Optional[float] = None
    v_identity: Optional[float] = None
    d_identity: Optional[float] = None
    j_identity: Optional[float] = None
    c_identity: Optional[float] = None
    v_sequence_start: Optional[int] = None
    v_sequence_end: Optional[int] = None
    d_sequence_start: Optional[int] = None
    d_sequence_end: Optional[int] = None
    j_sequence_start: Optional[int] = None
    j_sequence_end: Optional[int] = None
    c_sequence_start: Optional[int] = None
    c_sequence_end: Optional[int] = None

    v_germline_start: Optional[int] = None
    v_germline_end: Optional[int] = None
    d_germline_start: Optional[int] = None
    d_germline_end: Optional[int] = None
    j_germline_start: Optional[int] = None
    j_germline_end: Optional[int] = None
    c_germline_start: Optional[int] = None
    c_germline_end: Optional[int] = None

    fwr1_start: Optional[int] = None
    fwr1_end: Optional[int] = None
    cdr1_start: Optional[int] = None
    cdr1_end: Optional[int] = None
    fwr2_start: Optional[int] = None
    fwr2_end: Optional[int] = None
    cdr2_start: Optional[int] = None
    cdr2_end: Optional[int] = None
    fwr3_start: Optional[int] = None
    fwr3_end: Optional[int] = None
    cdr3_start: Optional[int] = None
    cdr3_end: Optional[int] = None
    fwr4_start: Optional[int] = None
    fwr4_end: Optional[int] = None

    sequence_aa_scheme_cigar: Optional[Cigar] = None
    scheme_residue_mapping: Optional[dict[str, str]] = None
    positional_scheme_mapping: Optional[dict[int, str]] = None
    exc: Optional[str] = None
    additional_validation_flags: Optional[EntryValidationResults] = None


@dataclass
class AirrRearrangementEntryAA:  # pylint: disable=too-many-instance-attributes
    """https://docs.airr-community.org/en/stable/datarep/rearrangements.html"""

    sequence_header: str
    sequence_aa: str
    numbering_scheme: str
    locus: Optional[str] = None
    stop_codon: Optional[bool] = None
    productive: Optional[bool] = None  # V and J aligned and no stop codon
    complete_vdj: Optional[bool] = None
    v_call: Optional[str] = None
    j_call: Optional[str] = None

    germline_alignment_aa: Optional[str] = None
    sequence_alignment_aa: Optional[str] = None
    v_alignment_start_aa: Optional[int] = None
    v_alignment_end_aa: Optional[int] = None
    j_alignment_start_aa: Optional[int] = None
    j_alignment_end_aa: Optional[int] = None
    v_sequence_alignment_aa: Optional[str] = None
    v_germline_alignment_aa: Optional[str] = None
    j_sequence_alignment_aa: Optional[str] = None
    j_germline_alignment_aa: Optional[str] = None

    fwr1_aa: Optional[str] = None
    cdr1_aa: Optional[str] = None
    fwr2_aa: Optional[str] = None
    cdr2_aa: Optional[str] = None
    fwr3_aa: Optional[str] = None
    cdr3_aa: Optional[str] = None
    fwr4_aa: Optional[str] = None
    junction_aa: Optional[str] = None
    junction_aa_length: Optional[int] = None

    v_score_aa: Optional[float] = None
    j_score_aa: Optional[float] = None

    v_cigar_aa: Optional[str] = None
    j_cigar_aa: Optional[str] = None

    v_support_aa: Optional[float] = None
    j_support_aa: Optional[float] = None
    v_identity_aa: Optional[float] = None
    j_identity_aa: Optional[float] = None

    v_sequence_start_aa: Optional[int] = None
    v_sequence_end_aa: Optional[int] = None
    j_sequence_start_aa: Optional[int] = None
    j_sequence_end_aa: Optional[int] = None

    v_germline_start_aa: Optional[int] = None
    v_germline_end_aa: Optional[int] = None
    j_germline_start_aa: Optional[int] = None
    j_germline_end_aa: Optional[int] = None

    fwr1_start_aa: Optional[int] = None
    fwr1_end_aa: Optional[int] = None
    cdr1_start_aa: Optional[int] = None
    cdr1_end_aa: Optional[int] = None
    fwr2_start_aa: Optional[int] = None
    fwr2_end_aa: Optional[int] = None
    cdr2_start_aa: Optional[int] = None
    cdr2_end_aa: Optional[int] = None
    fwr3_start_aa: Optional[int] = None
    fwr3_end_aa: Optional[int] = None
    cdr3_start_aa: Optional[int] = None
    cdr3_end_aa: Optional[int] = None
    fwr4_start_aa: Optional[int] = None
    fwr4_end_aa: Optional[int] = None

    sequence_aa_scheme_cigar: Optional[Cigar] = None
    scheme_residue_mapping: Optional[dict[str, str]] = None
    positional_scheme_mapping: Optional[dict[int, str]] = None
    exc: Optional[str] = None
    additional_validation_flags: Optional[EntryValidationResults] = None


AirrRearrangementEntry_co = TypeVar(
    "AirrRearrangementEntry_co", AirrRearrangementEntryNT, AirrRearrangementEntryAA, covariant=True
)


def serialize_airr_entry(entry: AirrRearrangementEntry_co) -> dict[str, str]:
    entry_dict = entry.__dict__
    entry_dict["scheme_residue_mapping"] = entry_dict["scheme_residue_mapping"] and base64_encode(
        json.dumps(entry_dict["scheme_residue_mapping"], separators=(",", ":"))
    )
    entry_dict["positional_scheme_mapping"] = entry_dict["positional_scheme_mapping"] and base64_encode(
        json.dumps(entry_dict["positional_scheme_mapping"], separators=(",", ":"))
    )
    entry_dict["exc"] = entry_dict["exc"] and base64_encode(entry_dict["exc"])
    entry_dict["additional_validation_flags"] = entry_dict["additional_validation_flags"] and base64_encode(
        json.dumps(entry_dict["additional_validation_flags"], separators=(",", ":"))
    )

    return entry_dict


@dataclass
class RegionOffsetsAA:  # pylint: disable=too-many-instance-attributes
    fwr1_start_aa: int
    fwr1_end_aa: int
    cdr1_start_aa: int
    cdr1_end_aa: int
    fwr2_start_aa: int
    fwr2_end_aa: int
    cdr2_start_aa: int
    cdr2_end_aa: int
    fwr3_start_aa: int
    fwr3_end_aa: int
    cdr3_start_aa: int
    cdr3_end_aa: int
    fwr4_start_aa: int
    fwr4_end_aa: int


@dataclass
class RegionOffsetsNT:  # pylint: disable=too-many-instance-attributes
    fwr1_start: int
    fwr1_end: int
    cdr1_start: int
    cdr1_end: int
    fwr2_start: int
    fwr2_end: int
    cdr2_start: int
    cdr2_end: int
    fwr3_start: int
    fwr3_end: int
    cdr3_start: int
    cdr3_end: int
    fwr4_start: int
    fwr4_end: int


GeneId = str
GeneSequence = str


@dataclass(frozen=True)
class Gene:
    name: GeneId
    species: Organism
    locus: Locus
    sequence: GeneSequence
    reading_frame: Optional[int]


@dataclass(frozen=True)
class GeneAA:
    name: GeneId
    species: Organism
    locus: Locus
    sequence: GeneSequence


@dataclass(frozen=True)
class SpeciesGeneMatch:
    gene_id: str
    rev_comp: bool
    coverage: int
    species: Organism
    locus: Locus


@dataclass(frozen=True)
class SpeciesPrefilteringResult:
    query: str
    rev_comp_query: str
    top_matches: list[SpeciesGeneMatch]


@dataclass(frozen=True)
class AlignmentEntryNT:  # pylint: disable=too-many-instance-attributes
    target_id: str
    alignment_score: float
    seq_identity: float
    e_value: float
    q_start: int
    q_end: int
    q_len: int
    t_start: int
    t_end: int
    t_len: int
    cigar: str
    rev_comp: bool

    species: Organism
    locus: Locus
    q_seq: GeneSequence
    t_seq: GeneSequence
    reading_frame: Optional[int]

    def __lt__(self, other):
        if self.e_value == other.e_value:
            self_aln_len = self.t_end - self.t_start
            other_aln_len = self.t_end - self.t_start
            if self_aln_len == other_aln_len:
                return self.target_id < other.target_id
            return self_aln_len > other_aln_len
        return self.e_value < other.e_value

    def __gt__(self, other):
        if self.e_value == other.e_value:
            self_aln_len = self.t_end - self.t_start
            other_aln_len = self.t_end - self.t_start
            if self_aln_len == other_aln_len:
                return self.target_id > other.target_id
            return self_aln_len < other_aln_len
        return self.e_value > other.e_value


@dataclass(frozen=True)
class AlignmentEntryAA:  # pylint: disable=too-many-instance-attributes
    target_id: str
    alignment_score: Optional[float]
    seq_identity: Optional[float]
    e_value: Optional[float]
    q_start: int
    q_end: int
    t_start: int
    t_end: int
    cigar: Cigar

    species: Organism
    locus: Locus
    q_seq: GeneSequence
    t_seq: GeneSequence

    def __lt__(self, other):
        if self.e_value == other.e_value:
            self_aln_len = self.t_end - self.t_start
            other_aln_len = self.t_end - self.t_start
            if self_aln_len == other_aln_len:
                return self.target_id < other.target_id
            return self_aln_len > other_aln_len
        return self.e_value < other.e_value

    def __gt__(self, other):
        if self.e_value == other.e_value:
            self_aln_len = self.t_end - self.t_start
            other_aln_len = self.t_end - self.t_start
            if self_aln_len == other_aln_len:
                return self.target_id > other.target_id
            return self_aln_len < other_aln_len
        return self.e_value > other.e_value


AlignmentEntry = TypeVar("AlignmentEntry", AlignmentEntryNT, AlignmentEntryAA)


@dataclass
class AlignmentsNT:
    v: Optional[AlignmentEntryNT] = None
    d: Optional[AlignmentEntryNT] = None
    j: Optional[AlignmentEntryNT] = None
    c: Optional[AlignmentEntryNT] = None


@dataclass
class AlignmentsAA:
    v: Optional[AlignmentEntryAA] = None
    j: Optional[AlignmentEntryAA] = None


@dataclass
class TranslatedAlignmentsNT:
    translated_query: str
    reading_frame: int
    aa_alignments: AlignmentsAA


class GermlineGene(Enum):
    V = "V"
    D = "D"
    J = "J"
    C = "C"


class InputType(str, Enum):
    NT = "nt"
    AA = "aa"



import time
from pathlib import Path
from typing import Any, Optional

import click
import psutil

from riot_na.api.api_mp import run_on_file_mp
from riot_na.api.riot_numbering import create_riot_aa, create_riot_nt
from riot_na.common.io import write_airr_iter_to_csv
from riot_na.config import GENE_DB_DIR
from riot_na.data.model import (
    AirrRearrangementEntryAA,
    AirrRearrangementEntryNT,
    InputType,
    Organism,
    Scheme,
)


@click.command()
@click.option("-f", "--input-file", type=Path, help="Path to input FASTA file.")
@click.option("-s", "--sequence", type=str, help="Input sequence file.")
@click.option("-o", "--output-file", type=Path, help="Path to output CSV file. If not specified, stdout is used.")
@click.option(
    "--scheme",
    type=click.Choice(Scheme),  # type: ignore
    default=Scheme.IMGT,
    help="Which numbering scheme should be used: imgt, kabat, chothia, martin. Default IMGT",
)
@click.option(
    "--species",
    type=click.Choice(Organism),  # type: ignore
    default=None,
    help="Which species germline sequences should be used. Default is all species.",
)
@click.option(
    "--input-type",
    type=click.Choice(InputType),  # type: ignore
    default=InputType.NT,
    help="What kind of sequences are provided on input. Default is nucleotide sequences.",
)
@click.option(
    "-p",
    "--ncpu",
    type=int,
    default=psutil.cpu_count(logical=False),
    help="Number of parallel processes to use. Default is number of physical cores.",
)
@click.option(
    "-e",
    "--extend_alignment",
    type=bool,
    default=False,
    help=(
        "Include unaligned beginning of the query sequence in numbering."
        "This option impacts only amino acid sequences."
    ),
)
def run_riot(
    input_file: Optional[Path],
    sequence: Optional[str],
    output_file: Optional[Path],
    scheme: Scheme,
    species: Optional[Organism],
    input_type: InputType,
    ncpu: int,
    extend_alignment: bool,
):
    species_list = [species] if species else None
    if input_file and input_file.exists():
        if not output_file:
            output_file = Path() / "numbering_result.csv"
        start = time.perf_counter()
        run_on_file_mp(
            GENE_DB_DIR,
            input_file,
            output_file,
            input_format="fasta",
            scheme=scheme,
            allowed_species=species_list,
            n_processes=ncpu,
            input_type=input_type,
        )
        end = time.perf_counter()
        elapsed_time = end - start
        print("Execution time:", time.strftime("%H:%M:%S", time.gmtime(elapsed_time)))
    elif sequence:
        result: Any
        record_type: Any
        match input_type:
            case InputType.NT:
                numbering_nt = create_riot_nt(allowed_species=species_list, db_dir=GENE_DB_DIR)
                record_type = AirrRearrangementEntryNT
                result = numbering_nt.run_on_sequence(header="-", query_sequence=sequence, scheme=scheme)
            case InputType.AA:
                numbering_aa = create_riot_aa(allowed_species=species_list, db_dir=GENE_DB_DIR)
                record_type = AirrRearrangementEntryAA
                result = numbering_aa.run_on_sequence(
                    header="-", query_sequence=sequence, scheme=scheme, extend_alignment=extend_alignment
                )

        if output_file:
            write_airr_iter_to_csv(output_file, record_type, [result])
        else:
            print(result.__dict__)

    else:
        print("Need to specify input sequence or FASTA file!")
